searchNodes=[{"doc":"Just-In-Time nif generator, FFI generator, C-compiler based on TinyCC. For Linux, MacOS, Windows (msys2) Using Niffler Once installed, you can use the Niffler to define new nif modules using embedded C fragments: defmodule Example do use Niffler defnif :count_zeros , [ str : :binary ] , [ ret : :int ] do &quot;&quot;&quot; while($str.size--) { if (*$str.data++ == 0) $ret++; } &quot;&quot;&quot; end end { :ok , [ 2 ] } = Example . count_zeros ( &lt;&lt; 0 , 1 , 0 &gt;&gt; ) See Niffler.defnif/4 for a full explanation of the parameters passed. Variable binding in C fragments Each c fragment is wrapped in shortcut macros to receive direct access to the defined parameters. So when defining a nif: defnif :count_zeros , [ str : :binary ] , [ ret : :int ] do # ^name ^input ^output There will be macros defined for each defined input and output variable. Each macro is prefixed with a dollar sign $ to highlight the fact that it's a macro and not real variable: In the exampe above: $str will be defined as a macro alias pointing to a binary type $ret will be defined as a macro alias to an integer Input and output variables are allocated on the stack of the nif function call making them thread-safe and isolated. Working with binaries When using binaries as input or output they are returned as as a struct with two members: typedef struct { uint64_t size; unsigned char* data; } Binary; The size and data fields can be used to read from inputs and write to outputs. Warning: NEVER write to input binaries. These are pointers into the BEAM VM, changing their values will have unknown but likely horrible consequences. Constructing output binaries requires care. The easiest way is to use the built-in macro function $alloc(size_t) which allows to allocate memory temporary during the runtime of the nif, that will be automatically freed. Other possibilities are to use the system malloc() directly but then free() needs to be called at a later point in time, or to use static memory in the module. Stack variables (or from alloca ) don't work as they are being destroyed before the Niffler program returns and the result values are beeing read. Two examples that are possible here: defmodule BinaryExample use Niffler defnif :static , [ ] , [ ret : :binary ] do &quot;&quot;&quot; static char str[10]; $ret.size = sizeof(str); $ret.data = str; for (int i = 0; i &lt; sizeof(str); i++) str[i] = i; &quot;&quot;&quot; end defnif :reverse , [ input : :binary ] , [ ret : :binary ] do &quot;&quot;&quot; $ret.data = $alloc($input.size); $ret.size = $input.size; for (int i = 0; i &lt; $input.size; i++) $ret.data[i] = $input.data[$input.size-(i+1)]; &quot;&quot;&quot; end end { :ok , [ &lt;&lt; 0 , 1 , 2 , 3 , 4 , 5 , 6 , 7 , 8 , 9 &gt;&gt; ] } = BinaryExample . static ( ) { :ok , [ &lt;&lt; 3 , 2 , 1 &gt;&gt; ] } = BinaryExample . reverse ( &lt;&lt; 1 , 2 , 3 &gt;&gt; ) Concurrency Each generated Niffler function is a small c program in its own call stack. Multiple runs of the same program are all executed in the same context. This allows to keep state in c programs when needed but also increases the chances for concurrency issues. Lets take this stateful counter: defnif :counter , [ ] , [ ret : :int ] do &quot;&quot;&quot; static uint64_t counter = 0; $ret = counter++; &quot;&quot;&quot; end The niffler protects against certain types concurrency issues by creating all input and output variables on the call stack. Two functions calls at the same time do never interefere on the input and output parameters. But this protection is not true for static and global variables. The above counter example is using a static counter variable, this variable will be the same instance on each call. So with high concurrency these things could happen: counter returns the same value for two concurrent calls. counter skips a value for two concurrent calls. The same problem affects the static binary example above. When called multiple times concurrently it will overwrite the static variable multiple times return undefined results. Defining helper functions When using Niffler.defnif/4 you sometimes might want to create helper functions or defines outside the function body. For short fragements it's possible to use the DO_RUN and END_RUN macros to separate the function body from global helpers: Here an example defining a recursive fibonacci function. In order to refer to the function name fib() recursively in c it needs to be defined. So we define it globally outside of an explicity DO_RUN / END_RUN block: defnif :fib , [ a : :int ] , ret : :int do &quot;&quot;&quot; int64_t fib(int64_t f) { if (f &lt; 2) return 1; return fib(f-1) + fib(f-2); } DO_RUN $ret = fib($a); END_RUN &quot;&quot;&quot; end Interally DO_RUN and END_RUN are c-macros that will be converted to the correct niffler wrapping to execute the code, while anything outside the DO_RUN / END_RUN block will be copied into the c code without modification. For larger blocks it might though be better to use Niffler.Library and override the Niffler.Library.c:header/0 callback. Using shared libraries (.dll, .so, .dylib) For working with shared library and create foreign function interfaces (FFI) for those please look at Niffler.Library Standard library Niffler comes with a minimal c standard library. Please check standard c documentation for reference. This is just a list of defined functions and types: / * types * / typedef signed char int8_t ; typedef unsigned char uint8_t ; typedef short int16_t ; typedef unsigned short uint16_t ; typedef int int32_t ; typedef unsigned uint32_t ; typedef long long int64_t ; typedef unsigned long long uint64_t ; typedef struct { uint64_t size ; unsigned char * data ; } Binary ; / * niffler helper * / void * $ alloc ( size_t size ) ; / * stdarg . h * / typedef __builtin_va_list va_list ; / * stddef . h * / typedef __SIZE_TYPE__ size_t ; typedef __PTRDIFF_TYPE__ ssize_t ; typedef __WCHAR_TYPE__ wchar_t ; typedef __PTRDIFF_TYPE__ ptrdiff_t ; typedef __PTRDIFF_TYPE__ intptr_t ; typedef __SIZE_TYPE__ uintptr_t ; void * alloca ( size_t size ) ; / * stdlib . h * / void * calloc ( size_t nmemb , size_t size ) ; void * malloc ( size_t size ) ; void free ( void * ptr ) ; void * realloc ( void * ptr , size_t size ) ; int atoi ( const char * nptr ) ; long int strtol ( const char * nptr , char * * endptr , int base ) ; unsigned long int strtoul ( const char * nptr , char * * endptr , int base ) ; void exit ( int ) ; / * stdio . h * / extern FILE * stdin ; extern FILE * stdout ; extern FILE * stderr ; FILE * fopen ( const char * path , const char * mode ) ; FILE * fdopen ( int fildes , const char * mode ) ; FILE * freopen ( const char * path , const char * mode , FILE * stream ) ; int fclose ( FILE * stream ) ; size_t fread ( void * ptr , size_t size , size_t nmemb , FILE * stream ) ; size_t fwrite ( void * ptr , size_t size , size_t nmemb , FILE * stream ) ; int fgetc ( FILE * stream ) ; int fputs ( const char * s , FILE * stream ) ; char * fgets ( char * s , int size , FILE * stream ) ; int getc ( FILE * stream ) ; int getchar ( void ) ; char * gets ( char * s ) ; int ungetc ( int c , FILE * stream ) ; int fflush ( FILE * stream ) ; int putchar ( int c ) ; int printf ( const char * format , ... ) ; int fprintf ( FILE * stream , const char * format , ... ) ; int sprintf ( char * str , const char * format , ... ) ; int snprintf ( char * str , size_t size , const char * format , ... ) ; int asprintf ( char * * strp , const char * format , ... ) ; int vprintf ( const char * format , va_list ap ) ; int vfprintf ( FILE * stream , const char * format , va_list ap ) ; int vsprintf ( char * str , const char * format , va_list ap ) ; int vsnprintf ( char * str , size_t size , const char * format , va_list ap ) ; int vasprintf ( char * * strp , const char * format , va_list ap ) ; void perror ( const char * s ) ; / * string . h * / char * strcat ( char * dest , const char * src ) ; char * strchr ( const char * s , int c ) ; char * strrchr ( const char * s , int c ) ; char * strcpy ( char * dest , const char * src ) ; void * memcpy ( void * dest , const void * src , size_t n ) ; void * memmove ( void * dest , const void * src , size_t n ) ; void * memset ( void * s , int c , size_t n ) ; char * strdup ( const char * s ) ; size_t strlen ( const char * s ) ; / * dlfcn . h * / void * dlopen ( const char * filename , int flag ) ; const char * dlerror ( void ) ; void * dlsym ( void * handle , char * symbol ) ; int dlclose ( void * handle ) ;","ref":"Niffler.html","title":"Niffler","type":"module"},{"doc":"Low level function takes a string as input and compiles it into a nif program. Returning the program reference. Prefer using the high-level function Niffler.defnif/4 or Niffler.Library instead. Examples iex&gt; { :ok , prog } = Niffler . compile ( &quot;$ret = $a * $b;&quot; , [ a : :int , b : :int ] , [ ret : :int ] ) iex&gt; Niffler . run ( prog , [ 3 , 4 ] ) { :ok , [ 12 ] } iex&gt; code = &quot;for (int i = 0; i &lt; $str.size; i++) if ($str.data[i] == 0) $ret++;&quot; iex&gt; { :ok , prog } = Niffler . compile ( code , [ str : :binary ] , [ ret : :int ] ) iex&gt; Niffler . run ( prog , [ &lt;&lt; 0 , 1 , 1 , 0 , 1 , 5 , 0 &gt;&gt; ] ) { :ok , [ 3 ] }","ref":"Niffler.html#compile/3","title":"Niffler.compile/3","type":"function"},{"doc":"Defines a new nif member method. To use defnif() import Niffler into your module with use Niffler . defnif takes three parameters and a c-fragment function body: name - an atom, the name of the to be defined nif function inputs - a keyword list of the format [name: type] outputs - a keyword list of the format [name: type] The inputs and outputs keyword lists take atom() as names and types. The parameter names can be freely choosen* the currently supported types are: int or int64 - a signed 64-bit integer uint64 - an unsigned 64-bit integer double - a double (floating point number) binary - an Elixir binary/string defmodule Example do use Niffler defnif :count_zeros , [ str : :binary ] , [ ret : :int ] do &quot;&quot;&quot; while($str.size--) { if (*$str.data++ == 0) $ret++; } &quot;&quot;&quot; end end { :ok , [ 2 ] } = Example . count_zeros ( &lt;&lt; 0 , 1 , 0 &gt;&gt; )","ref":"Niffler.html#defnif/4","title":"Niffler.defnif/4","type":"macro"},{"doc":"Low level function that executes the given Niffler program and returns any output values. Prefer using the high-level function Niffler.defnif/4 or Niffler.Library instead. Examples iex&gt; { :ok , prog } = Niffler . compile ( &quot;$ret = $a &lt;&lt; 2;&quot; , [ a : :int ] , [ ret : :int ] ) iex&gt; Niffler . run ( prog , [ 5 ] ) { :ok , [ 20 ] }","ref":"Niffler.html#run/3","title":"Niffler.run/3","type":"function"},{"doc":"Niffler.Library allows to wrap dynamically loaded libraries. And create foreign function interfaces (FFI) for them. This usually requires: Shared headers such for type definitions Some init code involving dlopen() Static variables/state that is part of the library Potential deinit code involving dlclose() Niffler.Library allows to produce modules at runtime doing all these things and exposing access to the c-functions of a dynamic library. defmodule Gmp do use Niffler.Library , thread_safe : false @impl true def header ( ) do &quot;&quot;&quot; // library handle void *gmp; // types typedef struct { int _mp_alloc; int _mp_size; void *_mp_d; } __mpz_struct; typedef __mpz_struct mpz_t[1]; void (*mpz_init)(mpz_t); // ... &quot;&quot;&quot; end @impl true def on_load ( ) do &quot;&quot;&quot; // loading the library with dlopen() gmp = dlopen(&quot;libgmp. \#{ library_suffix ( ) } &quot;, RTLD_LAZY); if (!gmp) { return &quot;could not load libgmp&quot;; } // loading symbols: dlerror(); if (!(mpz_init = dlsym(gmp, &quot;__gmpz_init&quot;))) { return dlerror(); } // other symbols ... &quot;&quot;&quot; end @impl true def on_destroy ( ) do &quot;&quot;&quot; if (!gmp) { return; } // unloading dlclose(gmp); &quot;&quot;&quot; end # definining one or more operations here... defnif :mul , [ a : :int , b : :int ] , ret : :int do &quot;&quot;&quot; mpz_set_si(ma, $a); mpz_set_si(mb, $b); mpz_mul(mc, ma, mb); $ret = mpz_get_si(mc); &quot;&quot;&quot; end end Once defined the module functions can be used via: { ok , [ result ] } = Gmp . mul ( 4 , 5 )","ref":"Niffler.Library.html","title":"Niffler.Library","type":"behaviour"},{"doc":"Defines a new nif function in the current module. Same as Niffler.defnif/4 but with access to the current module context.","ref":"Niffler.Library.html#defnif/4","title":"Niffler.Library.defnif/4","type":"macro"},{"doc":"Return a c-fragement of a common header for type definitions, static variables and other neccesary c code. Example: @impl true def header ( ) do &quot;&quot;&quot; // library handle void *gmp; &quot;&quot;&quot; end","ref":"Niffler.Library.html#c:header/0","title":"Niffler.Library.header/0","type":"callback"},{"doc":"Returns the current platforms default library suffix: dll on windows dylib on mac so on linux Useful for dlopen() code to load the correct library: @impl true def on_load ( ) do &quot;&quot;&quot; // loading the library with dlopen() gmp = dlopen(&quot;libgmp. \#{ library_suffix ( ) } &quot;, RTLD_LAZY); if (!gmp) { return &quot;could not load libgmp&quot;; } &quot;&quot;&quot; end","ref":"Niffler.Library.html#library_suffix/0","title":"Niffler.Library.library_suffix/0","type":"function"},{"doc":"Return a c-fragement that is called when the module is unloaded. Typicallly this fragment would contain a call to dlclose() for a dynamic library. Example: @impl true def on_destroy ( ) do &quot;&quot;&quot; if (!gmp) { return; } // unloading dlclose(gmp); &quot;&quot;&quot; end","ref":"Niffler.Library.html#c:on_destroy/0","title":"Niffler.Library.on_destroy/0","type":"callback"},{"doc":"Return a c-fragement that is called on the first call to the module. Typicall this fragment would contain a call to dlopen() when loading a dynamic library. This c-fragment should return a char* (a common string in c) when any error has occured. Example: @impl true def on_load ( ) do &quot;&quot;&quot; // loading the library with dlopen() gmp = dlopen(&quot;libgmp. \#{ library_suffix ( ) } &quot;, RTLD_LAZY); if (!gmp) { return &quot;could not load libgmp&quot;; } &quot;&quot;&quot; end","ref":"Niffler.Library.html#c:on_load/0","title":"Niffler.Library.on_load/0","type":"callback"}]